@@ -1,6 +1,6 @@
  /* ================================================================== *
   * FILE: tclreadline.c
-  * $Id: a81d2fca2303840fd4166b846c1babe8740f9f23 $
+  * $Id: 3515d5bc9214e3b47ae9ccf1e3fc2345ab939a30 $
   * ---
   * tclreadline -- gnu readline for tcl
   * https://github.com/flightaware/tclreadline/
@@ -25,12 +25,33 @@
 #  include <readline/history.h>
 #endif
 
+/* Check, if Tcl version supports Tcl_Size,
+ * which was introduced in Tcl 8.7 and 9.
+ */
+#ifndef TCL_SIZE_MAX
+#include <limits.h>
+#define TCL_SIZE_MAX INT_MAX
 
+#ifndef Tcl_Size
+typedef int Tcl_Size;
+#endif
+
+#define TCL_SIZE_MODIFIER ""
+#define Tcl_GetSizeIntFromObj Tcl_GetIntFromObj
+#endif
+
+/* TCL 9 does not define CONST anymore */
+#ifndef CONST
+#define CONST const
+#endif
+
+#ifdef EXTEND_LINE_BUFFER
 /*
  * this prototype may be missing
  * in readline.h
  */
 void rl_extend_line_buffer(int len);
+#endif
 
 #ifdef EXECUTING_MACRO_HACK
 /**
@@ -99,6 +120,7 @@ static char* tclrl_eof_string            = (char*) NULL;
 static char* tclrl_custom_completer      = (char*) NULL;
 static char* tclrl_last_line             = (char*) NULL;
 static int   tclrl_use_builtin_completer = 1;
+static int   tclrl_use_history_expansion = 1;
 static int   tclrl_history_length        = -1;
 Tcl_Interp*  tclrl_interp                = (Tcl_Interp*) NULL;
 
@@ -206,12 +228,13 @@ TclReadlineCmd(ClientData clientData, Tcl_Interp *interp, int objc,
         "read", "initialize", "write", "add", "complete",
         "customcompleter", "builtincompleter", "eofchar",
         "reset-terminal", "bell", "text", "update",
-        (char *) NULL
+        "historyexpansion", (char *) NULL
     };
     enum SubCmdIdx {
         TCLRL_READ, TCLRL_INITIALIZE, TCLRL_WRITE, TCLRL_ADD, TCLRL_COMPLETE,
         TCLRL_CUSTOMCOMPLETER, TCLRL_BUILTINCOMPLETER, TCLRL_EOFCHAR,
-        TCLRL_RESET_TERMINAL, TCLRL_BELL, TCLRL_TEXT, TCLRL_UPDATE
+        TCLRL_RESET_TERMINAL, TCLRL_BELL, TCLRL_TEXT, TCLRL_UPDATE,
+        TCLRL_HISTORYEXPANSION
     };
 
     Tcl_ResetResult(interp); /* clear the result space */
@@ -354,16 +377,16 @@ TclReadlineCmd(ClientData clientData, Tcl_Interp *interp, int objc,
                 Tcl_WrongNumArgs(interp, 2, objv, "?boolean?");
                 return TCL_ERROR;
             } else if (3 == objc) {
-                int bool = tclrl_use_builtin_completer;
+                int boolval = tclrl_use_builtin_completer;
                 if (TCL_OK != Tcl_GetBoolean(interp,
                                  Tcl_GetStringFromObj(objv[2], 0),
-                                 &bool)) {
+                                 &boolval)) {
                     Tcl_AppendResult(interp,
                         "wrong # args: should be a boolean value.",
                         (char*) NULL);
                     return TCL_ERROR;
                 } else {
-                    tclrl_use_builtin_completer = bool;
+                    tclrl_use_builtin_completer = boolval;
                 }
             }
             Tcl_AppendResult(interp, tclrl_use_builtin_completer ? "1" : "0",
@@ -434,7 +457,6 @@ TclReadlineCmd(ClientData clientData, Tcl_Interp *interp, int objc,
 
             break;
 
-
         case TCLRL_TEXT:
             if (objc != 2) {
                 Tcl_WrongNumArgs(interp, 2, objv, "");
@@ -446,6 +468,27 @@ TclReadlineCmd(ClientData clientData, Tcl_Interp *interp, int objc,
                Tcl_NewStringObj(rl_line_buffer ? rl_line_buffer : "", -1));
             break;
 
+        case TCLRL_HISTORYEXPANSION:
+            if (objc > 3) {
+                Tcl_WrongNumArgs(interp, 2, objv, "?boolean?");
+                return TCL_ERROR;
+            } else if (3 == objc) {
+                int boolval = tclrl_use_history_expansion;
+                if (TCL_OK != Tcl_GetBoolean(interp,
+                                 Tcl_GetStringFromObj(objv[2], 0),
+                                 &boolval)) {
+                    Tcl_AppendResult(interp,
+                        "wrong # args: should be a boolean value.",
+                        (char*) NULL);
+                    return TCL_ERROR;
+                } else {
+                    tclrl_use_history_expansion = boolval;
+                }
+            }
+            Tcl_AppendResult(interp, tclrl_use_history_expansion ? "1" : "0",
+                (char*) NULL);
+            break;
+
         default:
             goto BAD_COMMAND;
             /* NOTREACHED */
@@ -502,24 +545,32 @@ TclReadlineLineCompleteHandler(char* ptr)
          */
 
         char* expansion = (char*) NULL;
-        int status = history_expand(ptr, &expansion);
+        char* expand_output = (char*) NULL;
+        if (tclrl_use_history_expansion) {
+            int status = history_expand(ptr, &expand_output);
 
-        if (status >= 2) {
-            /* TODO: make this a valid tcl output */
-            printf("%s\n", expansion);
-            FREE(ptr);
-            FREE(expansion);
-            return;
-        } else if (status <= -1) {
-            Tcl_AppendResult
-                (tclrl_interp, "error in history expansion: ", expansion, "\n", (char*) NULL);
+            if (status >= 2) {
+                /* TODO: make this a valid tcl output */
+                printf("%s\n", expand_output);
+                FREE(ptr);
+                FREE(expand_output);
+                return;
+            } else if (status <= -1) {
+                Tcl_AppendResult
+                    (tclrl_interp, "error in history expansion: ", expand_output, "\n", (char*) NULL);
                 TclReadlineTerminate(TCL_ERROR);
-            FREE(ptr);
-            FREE(expansion);
-            return;
+                FREE(ptr);
+                FREE(expand_output);
+                return;
+            } else if (status == 0) {
+                expansion = ptr;
+            } else { /* status == 1 */
+                expansion = expand_output;
+            }
         } else {
-            Tcl_AppendResult(tclrl_interp, expansion, (char*) NULL);
+            expansion = ptr;
         }
+        Tcl_AppendResult(tclrl_interp, expansion, (char*) NULL);
 
     #ifdef EXECUTING_MACRO_NAME
         /**
@@ -549,7 +600,7 @@ TclReadlineLineCompleteHandler(char* ptr)
          */
         TclReadlineTerminate(LINE_COMPLETE);
         FREE(ptr);
-        FREE(expansion);
+        FREE(expand_output);
     }
 }
 
@@ -563,9 +614,15 @@ int
 Tclreadline_Init(Tcl_Interp *interp)
 {
     int status;
- #ifdef USE_TCL_STUBS
-     Tcl_InitStubs(interp, "8.6", 0);
+    /* Require 8.6 or later (9.0 also ok) */
+#ifdef USE_TCL_STUBS
+    if ( NULL == Tcl_InitStubs(interp, "8.6-", 0))
+#else
+    if (NULL == Tcl_PkgRequire(interp, "Tcl", "8.6-", 0))
 #endif
+    {
+        return TCL_ERROR;
+    }
     Tcl_CreateObjCommand(interp, "::tclreadline::readline", TclReadlineCmd,
         (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
     tclrl_interp = interp;
@@ -670,7 +727,11 @@ TclReadlineCompletion(char* text, int start, int end)
     int status;
     rl_completion_append_character = ' '; /* reset, just in case ... */
 
-    if (text && ('!' == text[0]
+    /* Only enable history expansion like '!!<TAB>' if the rl_extend_line_buffer
+     * function is available; e.g. libedit doesn't provide it, and alternative
+     * approaches to replace the line buffer don't give the desired behavior */
+#ifdef EXTEND_LINE_BUFFER
+    if (tclrl_use_history_expansion && text && ('!' == text[0]
             || (start && rl_line_buffer[start - 1] == '!' /* for '$' */))) {
         char* expansion = (char*) NULL;
         int oldlen = strlen(rl_line_buffer);
@@ -691,12 +752,13 @@ TclReadlineCompletion(char* text, int start, int end)
         }
         FREE(expansion);
     }
+#endif
 
     if (tclrl_custom_completer) {
         char start_s[BUFSIZ], end_s[BUFSIZ];
         Tcl_Obj* obj;
         Tcl_Obj** objv;
-        int objc;
+        Tcl_Size objc;
         int state;
         char* quoted_text = TclReadlineQuote(text, "$[]{}\"");
         char* quoted_rl_line_buffer = TclReadlineQuote(rl_line_buffer, "$[]{}\"");
@@ -723,7 +785,8 @@ TclReadlineCompletion(char* text, int start, int end)
             return matches;
 
         if (objc) {
-            int i, length;
+            int i;
+            Tcl_Size length;
             matches = (char**) MALLOC(sizeof(char*) * (objc + 1));
             for (i = 0; i < objc; i++) {
                 matches[i] = strdup(Tcl_GetStringFromObj(objv[i], &length));
@@ -893,4 +956,3 @@ TclReadlineParse(char** args, int maxargs, char* buf)
     *args = NULL;
     return nr;
 }
-
